<?php

namespace Drush\Commands\dev_modules;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ExtensionNameLengthException;
use Drupal\Core\Extension\MissingDependencyException;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drush\Attributes\Argument;
use Drush\Attributes\Command;
use Drush\Attributes\Option;
use Drush\Attributes\Usage;
use Drush\Commands\DrushCommands;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Global Drush command to enable/disable all development modules at once.
 */
class DevModulesCommands extends DrushCommands {

  /**
   * Flag indicating if dev modules should be enabled or disabled.
   *
   * @var bool
   */
  protected bool $enable = TRUE;

  /**
   * List of development modules.
   *
   * @var string[]
   */
  protected array $developmentModules;

  /**
   * List of non-development modules.
   *
   * @var string[]
   */
  protected array $productionModules;

  /**
   * List of modules that get enabled.
   *
   * @var string[]
   */
  protected array $enableModules;

  /**
   * List of modules that get disabled.
   *
   * @var string[]
   */
  protected array $disableModules;

  /**
   * DevModulesCommands constructor.
   */
  public function __construct(
    private readonly ConfigFactoryInterface $configFactory,
    private readonly ModuleHandlerInterface $moduleHandler,
    private readonly ModuleInstallerInterface $moduleInstaller,
  ) {
    parent::__construct();
    $this->developmentModules = $this->developmentDefaults();
    $this->productionModules = $this->productionDefaults();
    $this->alterModuleLists();
  }

  /**
   * Return an instance of these Drush commands.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The container.
   *
   * @return \Drush\Commands\dev_modules\DevModulesCommands
   *   The instance of Drush commands.
   */
  public static function create(ContainerInterface $container): DevModulesCommands {
    return new DevModulesCommands(
      $container->get('config.factory'),
      $container->get('module_handler'),
      $container->get('module_installer'),
    );
  }

  /**
   * Enable/disable all development modules at once.
   *
   * @throws \Exception
   */
  #[Command(name: 'site:dev', aliases: ['dev-modules', 'dev'])]
  #[Argument('flag', 'On or off, yes or no, True or False, 1 or 0.')]
  #[Option('list', 'Only determine the list of modules and print that list to the console.')]
  #[Usage(name: 'site:dev', description: 'Enable/disable all development modules at once.')]
  public function devMode(string $flag, array $options = ['list' => FALSE]): void {
    $list = $this
      ->setEnableMode($flag)
      ->prepareModules()
      ->processModules($options['list']);
    if ($options['list']) {
      $this->io()->section('Enable');
      $this->io()->listing($list['enable']);
      $this->io()->section('Disable');
      $this->io()->listing($list['disable']);
    }
  }

  /**
   * Init site by following instructions in given file.
   *
   * @throws \JsonException
   */
  #[Command(name: 'site:init', aliases: ['init-site'])]
  #[Argument('initFile', 'Filename with instructions.')]
  #[Usage(name: 'site:init filename', description: 'Init site by following instructions in given file.')]
  public function initSite(string $initFile): void {
    $instructions = json_decode(file_get_contents($initFile), TRUE, 512, JSON_THROW_ON_ERROR);
    foreach ($instructions['config'] as $configName => $keys) {
      $config = $this->configFactory->getEditable($configName);
      foreach ($keys as $key => $value) {
        $config->set($key, $value);
      }
      $config->save();
    }
  }

  /**
   * Invokes alter hooks to allow all modules to contribute to module lists.
   *
   * @return $this
   */
  protected function alterModuleLists(): DevModulesCommands {
    $this->developmentModules += $this->moduleHandler->invokeAll('dev_modules_dev');
    $this->productionModules += $this->moduleHandler->invokeAll('dev_modules_production');
    $this->moduleHandler->alter('dev_modules_dev', $this->developmentModules);
    $this->moduleHandler->alter('dev_modules_production', $this->productionModules);
    return $this;
  }

  /**
   * Determines if a given module exists.
   *
   * @param string $module
   *   Machine name of the module.
   *
   * @return bool
   *   TRUE if the module exists and is enabled, FALSE otherwise.
   */
  protected function moduleExists(string $module): bool {
    return $this->moduleHandler->moduleExists($module);
  }

  /**
   * Enables all the prepared modules.
   *
   * @return $this
   */
  protected function doEnableModules(): DevModulesCommands {
    try {
      $this->moduleInstaller->install($this->enableModules);
    }
    catch (ExtensionNameLengthException | MissingDependencyException) {
      // @todo Lof this exception.
    }
    return $this;
  }

  /**
   * Disables all the prepared modules.
   *
   * @return $this
   */
  protected function doDisableModules(): DevModulesCommands {
    $this->moduleInstaller->uninstall($this->disableModules);
    return $this;
  }

  /**
   * Callback to set the flag for subsequent operations.
   *
   * @param string $flag
   *   Flag to enable (on|yes|true|1) or disable (off|no|false|0) dev modules.
   *
   * @return $this
   *
   * @throws \Exception
   */
  public function setEnableMode(string $flag): DevModulesCommands {
    $this->enable = $this->flag($flag);
    return $this;
  }

  /**
   * Sanitizes the list of modules that need to be enabled.
   *
   * @param array $modules
   *   List of to be enabled modules.
   *
   * @return array
   *   Sanitized list of to be enabled modules.
   */
  protected function prepareEnableModule(array $modules): array {
    $module_data = $this->moduleHandler->getModuleList();
    $list = [];
    foreach ($modules as $module) {
      if (isset($module_data[$module]) && !$this->moduleExists($module)) {
        $list[] = $module;
      }
    }
    return $list;
  }

  /**
   * Sanitizes the list of modules that need to be disabled.
   *
   * @param array $modules
   *   List of to be disabled modules.
   *
   * @return array
   *   Sanitized list of to be disabled modules.
   */
  protected function prepareDisableModule(array $modules): array {
    $list = [];
    foreach ($modules as $module) {
      if ($this->moduleExists($module)) {
        $list[] = $module;
      }
    }
    return $list;
  }

  /**
   * Callback to prepare dev and prod modules.
   *
   * @return $this
   */
  public function prepareModules(): DevModulesCommands {
    if ($this->enable) {
      $this->enableModules = $this->prepareEnableModule($this->developmentModules);
      $this->disableModules = $this->prepareDisableModule($this->productionModules);
    }
    else {
      $this->enableModules = $this->prepareEnableModule($this->productionModules);
      $this->disableModules = $this->prepareDisableModule($this->developmentModules);
    }
    return $this;
  }

  /**
   * Callback to process the prepared lists of dev and prod modules.
   *
   * @param bool $listOnly
   *   Whether to only output the list of modules (TRUE) or really enable and
   *   disable them.
   *
   * @return array
   *   The lists containing the enabled and disabled modules.
   */
  public function processModules(bool $listOnly): array {
    if (!$listOnly) {
      $this
        ->doEnableModules()
        ->doDisableModules();
    }
    return [
      'enable' => $this->enableModules,
      'disable' => $this->disableModules,
    ];
  }

  /**
   * Returns default list of development modules.
   *
   * @return array
   *   List of development modules.
   */
  protected function developmentDefaults(): array {
    $modules = [
      'browser_refresh',
      'coffee',
      'devel',
      'hacked',
      'kint',
      'link_css',
      'rules_ui',
      'stage_file_proxy',
      'views_ui',
    ];
    if (!$this->moduleExists('layout_builder')) {
      $modules[] = 'field_ui';
    }
    return $modules;
  }

  /**
   * Returns default list of production modules.
   *
   * @return array
   *   List of production modules.
   */
  protected function productionDefaults(): array {
    return [
      'advagg',
      'bigpipe',
      'cdn',
      'cloudflare',
      'cloudflarepurger',
      'dynamic_page_cache',
      'page_cache',
      'purge',
      'varnish_purger',
    ];
  }

  /**
   * Validates the flag returns its boolean equivalent.
   *
   * @param string $flag
   *   Flag to enable (on|yes|true|1) or disable (off|no|false|0) dev modules.
   *
   * @return bool
   *   TRUE for valid positive flags, FALSE for valid negative flags.
   *
   * @throws \InvalidArgumentException
   */
  protected function flag(string $flag): bool {
    return match (strtolower($flag)) {
      'on', 'yes', 'true', '1' => TRUE,
      'off', 'no', 'false', '0' => FALSE,
      default => throw new \InvalidArgumentException('Unsupported flag.'),
    };
  }

}
